package api

import (
	"encoding/json"
	"fmt"
	"log"
	"net/http"
	"time"

	"gorm.io/gorm"
)

type Purchase struct {
	gorm.Model    `json:"-"`
	TransactionID uint    `json:"-" gorm:"column:transaction"`
	ProductCode   int     `json:"code" gorm:"column:product"`
	Product       Product `json:"product" gorm:"foreignKey:ProductCode;references:Code"`
	Price         int     `json:"price"`
	Amount        int     `json:"amount"`
}

func (a *api) AddPurchase(num int, w http.ResponseWriter, req *http.Request) {
	var purchase []Purchase
	err := json.NewDecoder(req.Body).Decode(&purchase)
	if err != nil {
		log.Printf("Can't create purchase: %v", err)
		w.WriteHeader(http.StatusInternalServerError)
		return
	}
	total := 0
	for i, p := range purchase {
		var product Product
		err = a.db.Where("code = ?", p.ProductCode).First(&product).Error
		if err != nil {
			log.Printf("Can't get product %d: %v", p.ProductCode, err)
			w.WriteHeader(http.StatusNotAcceptable)
			return
		}

		total += product.Price * p.Amount
		purchase[i].Price = product.Price
	}
	if total == 0 {
		log.Printf("Empty purchase (%d)", num)
		w.WriteHeader(http.StatusNotAcceptable)
		return
	}

	transaction := Transaction{
		MemberNum: num,
		Date:      time.Now(),
		Purchase:  purchase,
		Type:      "purchase",
		Total:     -total,
	}
	var httpStatus int
	err = a.db.Transaction(func(tx *gorm.DB) error {
		httpStatus, err = createTransaction(tx, &transaction)
		if err != nil {
			return err
		}

		for _, p := range purchase {
			err := tx.Model(&Product{}).
				Where("code = ?", p.ProductCode).
				Update("stock", gorm.Expr("stock - ?", p.Amount)).Error
			if err != nil {
				httpStatus = http.StatusInternalServerError
				return fmt.Errorf("Can't update product stock %d-%d: %v", p.ProductCode, p.Amount, err)
			}
		}
		return nil
	})
	if err != nil {
		log.Println(err)
		w.WriteHeader(httpStatus)
		return
	}

	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusCreated)
	err = json.NewEncoder(w).Encode(transaction)
	if err != nil {
		log.Printf("Can't encode added purchase: %v", err)
		w.WriteHeader(http.StatusInternalServerError)
	}

}
